// SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
// SPDX-License-Identifier: LGPL-3.0-or-later
#include <ikarus/utils/makeenum.hh>

#pragma once
namespace Ikarus {
MAKE_ENUM(DBCOption, Raw, Reduced, Full);
}